/*
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University.
 * Copyright (c) 1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 *      Utah $Hdr: bootstrap.c 1.5 94/12/15$
 */
/*
 * Bootstrap the various built-in servers.
 */
#include <mach_kdb.h>

#include <mach/port.h>
#include <mach/message.h>
#include <mach/vm_param.h>
#include <ipc/ipc_port.h>
#include <kern/host.h>
#include <kern/task.h>
#include <kern/thread.h>
#include <kern/current.h>
#include <device/device_port.h>
#include <sys/reboot.h>

mach_port_t	bootstrap_master_device_port;	/* local name */
mach_port_t	bootstrap_master_host_port;	/* local name */

int	boot_load_program();

extern char	*root_name;

char	*startup_name = "/mach_servers/startup";
char	 new_startup[64];
char	*emulator_name = "/mach_servers/emulator";

char	*startup_args;	/* server args passed at boot prompt */

extern void	default_pager();
extern void	default_pager_setup();
extern int	root_dev_node;

static void bootstrap();	/* forward */

task_t	bootstrap_task;

#define	BOOTSTRAP_MAP_SIZE	(4 * 1024 * 1024)

mach_port_t task_insert_send_right(task, port)
	task_t task;
	ipc_port_t port;
{
	mach_port_t name;

	for (name = 1;; name++) {
		kern_return_t kr;

		kr = mach_port_insert_right(task->itk_space, name,
					    port, MACH_MSG_TYPE_PORT_SEND);
		if (kr == KERN_SUCCESS)
			break;
		assert(kr == KERN_NAME_EXISTS);
	}

	return name;
}


void bootstrap_create()
{
	/*
	 * Create the bootstrap task.
	 */

	if ((boothowto>>RB_SHIFT) & RB_ASKNAME) {
		printf("bootstrap server? [%s] ", startup_name);
		safe_gets(new_startup, sizeof(new_startup));
		if (new_startup[0] == 'n' && new_startup[1] == '\0') {
			printf("Not starting bootstrap task.\n");
			return;
		}
		if (new_startup[0]) {
			startup_name = new_startup;
		}
	}
	bootstrap_task = kernel_task_create(TASK_NULL, BOOTSTRAP_MAP_SIZE);

	/*
	 * Insert send rights to the master host and device ports.
	 */
	bootstrap_master_host_port =
		task_insert_send_right(bootstrap_task,
			ipc_port_make_send(realhost.host_priv_self));

	bootstrap_master_device_port =
		task_insert_send_right(bootstrap_task,
			ipc_port_make_send(master_device_port));

	/*
	 * Start the bootstrap thread.
	 */
	printf("bootstrap_task %08x\n", bootstrap_task);
	(void) kernel_thread(bootstrap_task, bootstrap, 0);
}

/*
 * From this point on, the bootstrap task uses the external
 * versions of the kernel interface routines.
 */
#include <mach/task_special_ports.h>
#include <mach_user_internal.h>
#include <mach_port_internal.h>
#include <boot_printf.h>

extern void	boot_printf_init();
extern boolean_t        debug_kernel;

static void bootstrap()
{
	register kern_return_t	result;
	mach_port_t		user_task;
	mach_port_t		user_act;
	mach_port_t		bootstrap_port;

	mach_port_t		my_task;

	char			flag_string[12];

	extern int		root_fs_node;
	extern char		*export_paging;

	current_act()->task = bootstrap_task; /* XXX horrible kludge */

	my_task = mach_task_self();

	boot_printf_init();
	printf("bootstrap task\n");

	/*
	 * Allocate a port to pass initial server ports back and forth.
	 */
	result = mach_port_allocate(my_task, MACH_PORT_RIGHT_RECEIVE,
				    &bootstrap_port);
	assert(result == KERN_SUCCESS);

	result = mach_port_insert_right(my_task, bootstrap_port,
					bootstrap_port,
					MACH_MSG_TYPE_MAKE_SEND);
	assert(result == KERN_SUCCESS);

	result = task_set_bootstrap_port(my_task, bootstrap_port);
	assert(result == KERN_SUCCESS);

	/*
	 * Set up the temporary in-kernel default pager.
	 */
	default_pager_setup(bootstrap_master_host_port,
			    bootstrap_master_device_port,
			    root_name);

	/*
	 * Create the user task and thread to run the startup file.
	 */
	result = task_create(my_task, FALSE, &user_task);
	if (result != KERN_SUCCESS)
	    panic("task_create %d", result);

	result = thread_create(user_task, &user_act);
	if (result != KERN_SUCCESS)
	    panic("thread_create %d", result);

	/*
	 * Load the startup file.
	 * Pass it a command line of
	 * "startup -boot_flags root_name"
	 */
	get_boot_flags(flag_string);

	result = boot_load_program(bootstrap_master_host_port,
				   bootstrap_master_device_port,
				   user_task,
				   user_act,
				   root_name,
				   startup_args,
				   startup_name,
				   flag_string,
				   root_name,
				   (char *)0);

	if (result != 0)
	    panic("boot_load_program %d", result);

	/*
	 * Read emulator symbol table.
	 * Startup symbol table was read inside boot_load_program.
	 */
	read_emulator_symbols(bootstrap_master_host_port,
				bootstrap_master_device_port,
				   root_name,
				   emulator_name);

	/*
	 * Start up the thread
	 */
	result = thread_resume(user_act);
	if (result != KERN_SUCCESS)
	    panic("thread_resume %d", result);

	{
	    struct imsg {
		mach_msg_header_t	hdr;
		mach_msg_type_t		port_desc_1;
		mach_port_t		port_1;
		mach_msg_type_t		port_desc_2;
		mach_port_t		port_2;
	    } imsg;

	    /*
	     * Wait for the port-request message
	     */
	    while (mach_msg(&imsg.hdr, MACH_RCV_MSG,
			    0, sizeof imsg.hdr, bootstrap_port,
			    MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL)
						!= MACH_MSG_SUCCESS)
		continue;

	    /*
	     * Send back the host and device ports
	     */

	    imsg.hdr.msgh_bits = MACH_MSGH_BITS_COMPLEX |
		MACH_MSGH_BITS(MACH_MSGH_BITS_REMOTE(imsg.hdr.msgh_bits), 0);
	    /* msgh_size doesn't need to be initialized */
	    /* leave msgh_remote_port untouched */
	    imsg.hdr.msgh_local_port = MACH_PORT_NULL;
	    /* leave msgh_seqno untouched */
	    imsg.hdr.msgh_id += 100;	/* this is a reply msg */

	    imsg.port_desc_1.msgt_name = MACH_MSG_TYPE_COPY_SEND;
	    imsg.port_desc_1.msgt_size = 32;
	    imsg.port_desc_1.msgt_number = 1;
	    imsg.port_desc_1.msgt_inline = TRUE;
	    imsg.port_desc_1.msgt_longform = FALSE;
	    imsg.port_desc_1.msgt_deallocate = FALSE;
	    imsg.port_desc_1.msgt_unused = 0;

	    imsg.port_1	= bootstrap_master_host_port;

	    imsg.port_desc_2 = imsg.port_desc_1;

	    imsg.port_2	= bootstrap_master_device_port;

	    result = mach_msg(&imsg.hdr, MACH_SEND_MSG,
			      sizeof imsg, 0, MACH_PORT_NULL,
			      MACH_MSG_TIMEOUT_NONE, MACH_PORT_NULL);
	    if (result != MACH_MSG_SUCCESS)
		panic("mach_msg %x", result);
	}
	printf("booted\n");

	/*
	 * Become the default pager
	 */
	default_pager(bootstrap_master_host_port);
	/*NOTREACHED*/
}

/*
 * Parse the boot flags into an argument string.
 * Format as a standard flag argument: '-asd...'
 */

get_boot_flags(str)
	char	str[];	/* OUT */
{
	register char *cp;
	register int	bflag;
	cp = str;
	*cp++ = '-';

	bflag = boothowto;

	if (bflag & RB_ASKNAME)
	    *cp++ = 'a';
	if (bflag & RB_SINGLE)
	    *cp++ = 's';
#if	MACH_KDB
	if (bflag & RB_KDB)
	    *cp++ = 'd';
#endif	MACH_KDB
	if (bflag & RB_HALT)
	    *cp++ = 'h';
	if (bflag & RB_INITNAME)
	    *cp++ = 'n';
	if (bflag & RB_ALTBOOT)
	    *cp++ = 'v';

	if (cp == &str[1])	/* no flags */
	    *cp++ = 'x';
	*cp = '\0';
}
