/* 
 * Copyright (c) 1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: rd.c 1.8 92/06/27$
 *	Author: Bob Wheeler, University of Utah CSL
 */

/*
 *	CS80 disk driver for 27110B CIO HPIB controller
 */

#include <rd.h>
#if NRD > 0

#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#include <machine/machparam.h>
#include <machine/spl.h>

#include <hpcio/device.h>
#include <hpcio/hpibreg.h>
#include <hpcio/cs80.h>

#define	b_cylin	b_resid

extern caddr_t cio_malloc();
extern int dumpoffset, dumpsize;
extern dev_t dumpdev;

void	rdintr();

int	rdattach(), rdprobe(), rdscan(), rdlayer();
void	rddisplay();
struct	driver rddriver = {
	"rd", rdattach, rdprobe, rdscan, rddisplay, rdlayer
};

struct	rd_softc {
	int	unit;			/* rd unit number */
	int	hpibunit;		/* HPIB unit */
	int	slave;			/* HPIB slave number */
	int	flags;			/* flags */
	int	alive;			/* device is alive */
	struct	rdst *rdst;		/* disk configuration information */
} rd_softc[NRD];

struct hpib_request hpib_req[NRD];

/*
 * This is a mess. Some HP drives have an odd number of 256 byte sectors per 
 * track and an odd number of tracks per cylinder. To make matters worse HPUX 
 * (and hence hp800 BSD) defines some partition layouts on odd cylinders. We
 * can't win and so we will represent all partition tables in terms of 256
 * byte sectors. The kernel will still access the driver as if the block 
 * size was 512 and we'll deal with the conversion.
 */

/*
 * These maps support three basic layouts:
 *
 *	A/B/H:	 This is the "traditional" setup for a bootable disk.
 *		 A is the root partition, B the swap, and H a user partition,
 *		 G is the 2MB boot partition, and D is the 2MB kernel.
 *	A/B/E/F: Similar to A/B/D/H with E and F breaking H into two
 *		 partitions.  E could be used for /usr and F for users.
 *	C:	 This gives a single, non-bootable, large user filesystem.
 *		 Good for second drives on a machine (e.g. /usr/src).
 *
 * XXX - We should really be using disk labels
 */

struct	sizes {
	daddr_t	nsect;
	int	cyloff;
} hp2200A_sizes[8] = {
	97632,	9,		/* A=cyl 9 - 116    : 24 MB */
	131080,	117,		/* B=cyl 117 - 261  : 32 MB */
	1309896,0,		/* C=cyl 0 - 1448   : 335 MB */
	262160,	117,		/* D=cyl 117 - 406  : 64 MB */
	361600,	407,		/* E=cyl 407 - 806  : 90 MB */
	580368,	807,		/* F=cyl 807 - 1448 : 145 MB */
	1073048,262,		/* G=cyl 262 - 1448 : 268 MB */
	941968,	407		/* H=cyl 407 - 1448 : 235 MB */
}, hp2203A_sizes[8] = {
	97120,	5,		/* A=cyl   5 - 59   : 24 MB */
	74128,	60,		/* B=cyl  60 - 100  : 18 MB */
	2619792,0,		/* C=cyl   0 - 1448 : 670 MB */
	517088,	101,		/* D=cyl 101 - 386  : 132 MB */
	517088,	387,		/* E=cyl 387 - 672  : 132 MB */
	1403008,673,		/* F=cyl 673 - 1448 : 359 MB */
	7992,	0,		/* G=cyl   0 - 4    : 2 MB */
	2437184,101		/* H=cyl 101 - 1448 : 623 MB */
}, hp7937_sizes[8] = {
	95940,	6,		/* A=cyl 6 - 65     : 24  MB */
	195078,	66,		/* B=cyl 66 - 187   : 49  MB */
	2232204,0,		/* C=cyl 0 - 1395   : 558 MB */
	358176,	66,		/* D=cyl 66 - 289   : 90  MB */
	361374,	290,		/* E=cyl 290 - 515  : 90  MB */
	1407120,516,		/* F=cyl 516 - 1395 : 352 MB */
	1931592,188,		/* G=cyl 188 - 1395 : 483 MB */
	1768494,290,		/* H=cyl 290 - 1395 : 442 MB */
}, hp7959_sizes[8] = {
        97524,  11,             /* A=cyl 11 - 139   : 24 MB */
        131544, 140,            /* B=cyl 140 - 313  : 32 MB */
        1188432,0,              /* C=cyl 0 - 1571   : 304 MB */
        196560, 140,            /* D=cyl 140 - 399  : 48 MB */
        362880, 400,            /* E=cyl 400 - 879  : 90 MB */
        523152, 880,            /* F=cyl 880 - 1571 : 131 MB */
        951048, 314,            /* G=cyl 314 - 1571 : 238 MB */
        886032, 400             /* H=cyl 400 - 1571 : 221 MB */
};

struct rdst {
    char                *name;		/* drive name */
    int			nsect;	        /* sectors/track */
    int			ntrack;	        /* tracks/cylinder */
    int			nspc;	        /* sectors/cylinder */
    int			nrpm;	        /* revolutions/second */
    int			ncyl;	        /* cylinders */
    struct sizes	*sizes;		/* partition table */
    int			hwid;		/* hardware ID */
} rdst[] = {
    {"HP2200A", 113,  8,  113*8, 3600, 1449, hp2200A_sizes, 0x022f },
    {"HP2203A", 113, 16, 113*16, 3600, 1449, hp2203A_sizes, 0x0230 },
    {"HP7937",  123, 13, 123*13, 3600, 1396, hp7937_sizes,  0x0214 },
    {"HP7959",   63, 12,  63*12, 3600, 1572, hp7959_sizes,  0x022E },
    0
};

struct	buf rdtab[NRD];
struct	buf rdbuf[NRD];

#define	rdunit(x)	((minor(x) >> 3) & 0xf)
#define rdpart(x)	(minor(x) & 0x7)

rdprobe(hd)
	struct hpib_dev *hd;
{
	struct rdst *r;

	for (r = rdst; r->nsect; r++)
		if (hd->hwid == r->hwid)
			return 0;
	return ENXIO;
}

rdscan() 
{
	return 0;
}

rdattach(hd, unit, flags)
	struct hpib_dev *hd;
{
	struct rd_softc     *rs = &rd_softc[unit];
	struct hpib_request *hr = &hpib_req[unit];
	struct rdst *r;
	int status;

	for (r = rdst; r->nsect; r++) {
		if (hd->hwid != r->hwid) 
			continue;

		rs->unit     = unit;
		rs->hpibunit = hd->hpibunit;
		rs->slave    = hd->slave;
		rs->flags    = flags;
		rs->rdst     = r;

		/*
		 * allocate aligned space for the command buffer and the
		 * status.
		 */
		hr->command  = cio_malloc(HPIB_MIN_BUFFER, 
					  HPIB_BUFFER_ALIGN);
		hr->qstatus  = cio_malloc(HPIB_MIN_BUFFER, 
					  HPIB_BUFFER_ALIGN);
		hr->qstatcnt = sizeof(struct cs80_status);
		hr->unit     = unit;
		hr->hpibunit = hd->hpibunit;
		hr->done     = rdintr;

		rdtab[unit].b_actf = (struct buf *)0;
		rdtab[unit].b_active = 0;

		if ((status = rdreset(unit)) == 0)
			rs->alive = 1;
		else
			printf("rdattach: failed rdreset\n");

		/*
		 * Config the dump info for pdcdump.
		 */
		if (unit == rdunit(dumpdev)) {
			int part = rdpart(dumpdev);
			int size = r->sizes[part].nsect / 2;

			if (dumpsize > dtoc(size))
				dumpsize = dtoc(size);

			dumpoffset = (size - ctod(dumpsize)) +
				((r->nspc * r->sizes[part].cyloff) / 2);
		}
		return status;
	}

	return ENXIO;
}

void
rddisplay(unit)
{
}

rdlayer(unit, layer, devname)
	int *layer;
	char *devname;
{
	struct rd_softc *rs = &rd_softc[*layer];

	if (*layer >= NRD)
		return ENXIO;

	if (rs->alive) {
		*devname++ = 'r';
		*devname++ = 'd';
		*devname++ = unit % 10 + '0';
		*devname++ = 'a';
		return 0;
	}

	return ENXIO;
}


rdreset(unit)
{
	struct rd_softc     *rs = &rd_softc[unit];
	struct hpib_request *hr = &hpib_req[unit];
	char *p;
	int status;

	return 0;
}

rdopen(dev, flags)
	dev_t dev;
{
	register int unit = rdunit(dev);
	register struct rd_softc *rs = &rd_softc[unit];

	if (unit >= NRD || !rs->alive) {
		printf("rdopen: bad unit\n");
		return ENXIO;
	}

	return 0;
}

rdstrategy(bp)
	register struct buf *bp;
{
	register int unit = rdunit(bp->b_dev);
	register struct rd_softc *rs = &rd_softc[unit];
	register struct sizes *pinfo = &rs->rdst->sizes[rdpart(bp->b_dev)];
	register struct buf *dp = &rdtab[unit];
	register daddr_t bn;
	register int sz, s;

	bn = bp->b_blkno << 1;
	sz = (bp->b_bcount + 255) >> 8;
	if (bn < 0 || bn + sz > pinfo->nsect) {
		sz = pinfo->nsect - bn;
		if (sz == 0) {
			bp->b_resid = bp->b_bcount;
			iodone(bp);
			return;
		}
		if (sz < 0) {
			bp->b_error = EINVAL;
			bp->b_flags |= B_ERROR;
			iodone(bp);
			return;
		}
		bp->b_bcount = sz << 8;   /* XXX sz in sectors, not blocks */
	}
	bp->b_cylin = bn / rs->rdst->nspc + pinfo->cyloff;
	s = splbio();
	disksort(dp, bp);
	if (dp->b_active == 0) {
		dp->b_active = 1;
		rdustart(unit);
	}
	splx(s);
	return;
}

rdustart(unit)
{
       	struct buf          *bp;
	struct rd_softc     *rs = &rd_softc[unit];
	struct hpib_request *hr = &hpib_req[unit];
	char *p;
	int part;
	int address;

	while((bp = rdtab[unit].b_actf) != (struct buf *)0) {

		part = rdpart(bp->b_dev);
		address = (bp->b_blkno << 1) + 
			rs->rdst->nspc * rs->rdst->sizes[part].cyloff;
	
		/*
		 * build CS80 read or write command
		 */
		p = (char *) hr->command;
		if (bp->b_flags & B_READ) {
			hr->read = 1;
			cs80_read(p, rs->slave, 10);
			/*
			 * XXX For some reason, using any unit other than
			 * 0 causes errors. The device is actually specified
			 * in the HP specific part of the command bytes.
			 */
			set_unit(p, 0);		
			set_address(p, 0, address);
			set_length(p, bp->b_bcount);
			locate_and_read(p);
		} else {
			hr->read = 0;
			cs80_write(p, rs->slave, 10);
			set_unit(p, 0);
			set_address(p, 0, address);
			set_length(p, bp->b_bcount);
			locate_and_write(p);
		}
		hr->cmdcnt = p - (char *) hr->command;
		hr->buffer = bp->b_un.b_addr;
		hr->bufcnt = bp->b_bcount;

		bzero(hr->qstatus, sizeof(struct cs80_status));
		fdcache(0, hr->qstatus, sizeof(struct cs80_status));

		/*
		 * queue the HPIB request to the driver, if successful return
		 */
		if (!hpibgo(hr))
			return;

		/*
		 * there was an error queuing the last request, get rid of
		 * it and then look for another request
		 */
		printf("rdustart: hpipgo error: unit%d, part%d, address%d\n",
		       unit, part, address);
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		bp->b_resid = bp->b_bcount;
		rdtab[unit].b_actf = bp->b_actf;
		iodone(bp);
	}

	/*
	 * error with packet and nothing else to do
	 */
	rdtab[unit].b_active = 0;
}

void 
rdintr(unit)
{
	struct buf          *bp = rdtab[unit].b_actf;
	struct hpib_request *hr = &hpib_req[unit];
	int i;

	if (bp == (struct buf *) NULL) {
		printf("rd%d: bp == NULL\n", unit);
		return;
	}

	if (hr->status || hpib_error(hr->qstatus, 1) != 0) {
		printf("rdintr: unit%d, dev%d, count%d, blkno%d\n",
		       unit, bp->b_dev, bp->b_bcount, bp->b_blkno);
		for (i = 0; i < 16; i++)
			printf("%x ", hr->command[i]);
		printf("\n");
		bp->b_resid = bp->b_bcount;
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
	} else
		bp->b_resid = 0;

	rdtab[unit].b_actf = bp->b_actf;

	iodone(bp);
	if (rdtab[unit].b_actf) {
		rdustart(unit);
	} else
		rdtab[unit].b_active = 0;
}

rdread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register int unit = rdunit(dev);
	return physio(rdstrategy, &rdbuf[unit], dev, B_READ, minphys, uio);
}

rdwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register int unit = rdunit(dev);
	return physio(rdstrategy, &rdbuf[unit], dev, B_WRITE, minphys, uio);
}

rdioctl(dev, cmd, data, flag)
	dev_t dev;
	int cmd;
	caddr_t data;
	int flag;
{
	return EINVAL;
}

rdsize(dev)
	dev_t dev;
{
	int part = rdpart(dev);
	int unit = rdunit(dev);
	struct rd_softc *rs = &rd_softc[unit];

	if (unit >= NRD || !rs->alive)
		return -1;

	return rs->rdst->sizes[part].nsect / 2;
}

dump_bp(bp)
	struct buf *bp;
{
	printf("bp : %x\n", bp);
	printf("bp->b_bcount    = %x\n", bp->b_bcount);
	printf("bp->b_error     = %x\n", bp->b_error);
	printf("bp->b_dev       = %x\n", bp->b_dev);
	printf("bp->b_blkno     = %x\n", bp->b_blkno);
	printf("bp->b_resid     = %x\n", bp->b_resid);
	printf("bp->b_un.b_addr = %x\n\n", bp->b_un.b_addr);
}
#endif NRD > 0
